--- a/FL/Fl_Widget.H
+++ b/FL/Fl_Widget.H
@@ -842,6 +842,9 @@ public:
    */
   unsigned int visible_focus() { return flags_ & VISIBLE_FOCUS; }
 
+  static unsigned int fribidi();
+  static void fribidi(int v);
+
   /** The default callback for all widgets that don't set a callback.
 
     This callback function puts a pointer to the widget on the queue
--- a/src/Fl_Widget.cxx
+++ b/src/Fl_Widget.cxx
@@ -24,6 +24,69 @@
 #include <stdlib.h>
 #include "flstring.h"
 
+#if defined(FLTK_USE_FRIBIDI)
+
+#include <fribidi.h>
+#define FRIBIDI_MAX_STRLEN 65000
+
+static unsigned int fribidi_ = 0;
+
+unsigned int Fl_Widget::fribidi(void) {
+  return fribidi_;
+}
+
+void Fl_Widget::fribidi(int v) {
+  fribidi_ = v ? 1 : 0;
+}
+
+/** Returns pointer to allocated string or NULL on error. */
+static char *fribidi_parse_line(const char *input)
+{
+  char buffer[FRIBIDI_MAX_STRLEN];
+  size_t size;
+  FriBidiParType base = FRIBIDI_PAR_LTR;
+  FriBidiStrIndex len, *ltov = NULL, *vtol = NULL;
+  FriBidiChar logical[FRIBIDI_MAX_STRLEN];
+  FriBidiChar visual[FRIBIDI_MAX_STRLEN];
+  FriBidiLevel *levels = NULL;
+  FriBidiCharSet charset;
+
+  if (!input || (size = strlen(input)) == 0)
+    return NULL;
+
+  charset = fribidi_parse_charset("UTF-8");
+  fribidi_set_mirroring(true);
+  fribidi_set_reorder_nsm(false);
+
+  if (size >= FRIBIDI_MAX_STRLEN) {
+    size = FRIBIDI_MAX_STRLEN - 1;
+  }
+
+  strncpy(buffer, input, size);
+  len = fribidi_charset_to_unicode(charset, buffer, size, logical);
+
+  if (len == 0)
+    return NULL;
+
+  if (fribidi_log2vis(logical, len, &base, visual, ltov, vtol, levels)) {
+    len = fribidi_remove_bidi_marks(visual, len, ltov, vtol, levels);
+    if (len > 0) {
+      fribidi_unicode_to_charset(charset, visual, len, buffer);
+      return strdup(buffer);
+    }
+  }
+
+  return NULL;
+}
+
+#else
+
+/* no fribidi support */
+unsigned int Fl_Widget::fribidi(void) { return 0; }
+void Fl_Widget::fribidi(int) { }
+
+#endif
+
 
 ////////////////////////////////////////////////////////////////
 // for compatibility with Forms, all widgets without callbacks are
@@ -111,7 +174,17 @@ Fl_Widget::Fl_Widget(int X, int Y, int W, int H, const char* L) {
 
   x_ = X; y_ = Y; w_ = W; h_ = H;
 
-  label_.value	 = L;
+#if defined(FLTK_USE_FRIBIDI)
+  char *str;
+  if (L && fribidi() && (str = fribidi_parse_line(L))) {
+    label_.value = str;
+    set_flag(COPIED_LABEL);
+  } else {
+    label_.value = L;
+  }
+#else
+  label_.value   = L;
+#endif
   label_.image   = 0;
   label_.deimage = 0;
   label_.type	 = FL_NORMAL_LABEL;
@@ -283,7 +356,18 @@ void Fl_Widget::label(const char *a) {
     free((void *)(label_.value));
     clear_flag(COPIED_LABEL);
   }
+
+#if defined(FLTK_USE_FRIBIDI)
+  char *b = NULL;
+  if (fribidi() && (b = fribidi_parse_line(a))) {
+    label_.value=b;
+    set_flag(COPIED_LABEL);
+  } else {
+    label_.value=a;
+  }
+#else
   label_.value=a;
+#endif
   redraw_label();
 }
 
